package com.ijchart.xychart.annotation ;

import java.io.Serializable ;
import java.util.Arrays ;

import javax.swing.event.EventListenerList ;

import com.ijchart.xychart.event.AnnotationChangeEvent ;
import com.ijchart.xychart.listener.IAnnotationChangeListener ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;

/**
 * <p>Title: IJChart</p>
 * <zh>IAnnotationĳʵ.</zh>
 * <en>A base class for handling the drawing of {@link IAnnotation}</en>
 * <p>Description: a chart library for the Java(tm) platform.</p>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public abstract class AbstractAnnotation implements IAnnotation, Cloneable, Serializable,
  IAnnotationChangeListener {

  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -1305605489266328820L ;

  /**
   * <zh>,ɱл.</zh>
   * <en>The listener collection,not serializable.</en>
   */
  private transient EventListenerList listenerList ;

  /**
   * <zh>annotationʾַ.</zh>
   * <en>The tooltip text for annotation.</en>
   */
  private String tooltipText ;

  /**
   * <zh>ûвĹ.</zh>
   * <en>No argument constructor.</en>
   */
  public AbstractAnnotation() {
    this.listenerList = new EventListenerList() ;
    this.tooltipText = null ;
  }

  /**
   * <zh>annotationʾַ.</zh>
   * <en>Returns the tooltip text for annotation.</en>
   *
   * @return String <zh>ַ (Ϊnull).</zh>
   * <en>The tooltip text (maybe null).</en>
   */
  public String getTooltipText() {
    return this.tooltipText ;
  }

  /**
   * <zh>annotationʾַ,м
   * {@link com.ijchart.xychart.event.AnnotationChangeEvent}¼.</zh>
   * <en>Sets the tooltip text for annotation and sends an
   * {@link com.ijchart.xychart.event.AnnotationChangeEvent} to all registered listeners.</en>
   *
   * @param tooltipText String <zh>ַ (Ϊnull).</zh>
   * <en>The tooltip text (null permitted)..</en>
   */
  public void setTooltipText(String tooltipText) {
    if (!EqualsUtil.equals(this.tooltipText, tooltipText)) {
      this.tooltipText = tooltipText ;
      fireAnnotationChanged() ;
    }
  }

  /**
   * <zh>յע(annotation)仯֪ͨ.</zh>
   * <en>Receives notification of an {@link com.ijchart.xychart.annotation.IAnnotation}
   *  change event.</en>
   *
   * @param event <zh>ע(annotation)仯Ķ</zh>
   * <en>Information about the event.</en>
   */
  public void annotationChanged(AnnotationChangeEvent event) {
    notifyListeners(event) ;
  }

  /**
   * <zh>עע(annotation)仯ļ߶.</zh>
   * <en>Registers an object for notification of changes to the
   *  {@link com.ijchart.xychart.annotation.IAnnotation}</en>
   *
   * @param listener <zh>עļ.</zh>
   * <en>the object to register.</en>
   */
  public void addChangeListener(IAnnotationChangeListener listener) {
    listenerList.add(IAnnotationChangeListener.class, listener) ;
  }

  /**
   * <zh>ɾעע(annotation)仯߶.</zh>
   * <en>Deregisters an object for notification of changes to the
   *  {@link com.ijchart.xychart.annotation.IAnnotation}.</en>
   *
   * @param listener <zh>Ҫɾļ߶</zh>
   * <en>the object to deregister.</en>
   */
  public void removeChangeListener(IAnnotationChangeListener listener) {
    listenerList.remove(IAnnotationChangeListener.class, listener) ;
  }

  /**
   * <zh>еע(annotation)仯߶.</zh>
   * <en>Deregisters all object for notification of changes to the
   *  {@link com.ijchart.xychart.annotation.IAnnotation}.</en>
   */
  public void removeAllChangeListener() {
    Object[] listeners = this.listenerList.getListenerList() ;

    for (int i = listeners.length - 2 ; i >= 0 ; i -= 2) {
      this.listenerList.remove(IAnnotationChangeListener.class,
                               (IAnnotationChangeListener) listeners[i + 1]) ;
    }
  }

  /**
   * <zh>ָlistenerǷ,ҪڵԪ.</zh>
   * <en>Returns true if the specified object is registered with
   * the {@link com.ijchart.xychart.annotation.IAnnotation} as a listener.
   * Most applications won't need to call this
   * method, it exists mainly for use by unit testing code.</en>
   *
   * @param listener <zh>ҪԵļ.</zh>
   * <en>the listener.</en>
   * @return A boolean.
   */
  public boolean hasListener(IAnnotationChangeListener listener) {
    return Arrays.asList(this.listenerList.getListenerList()).contains(listener) ;
  }

  /**
   * <zh>ͨмע(annotation)˱仯.</zh>
   * <en>Notifies all registered listeners that the annotation has changed.</en>
   */
  protected void fireAnnotationChanged() {
    notifyListeners(new AnnotationChangeEvent(this)) ;
  }

  /**
   * <zh>֪ͨעļע(annotation)˱仯.</zh>
   * <en>Notifies all registered listeners that the annotation has changed.</en>
   *
   * @param event <zh>¼.</zh>
   * <en>contains information about the event that triggered the notification.</en>
   */
  protected void notifyListeners(AnnotationChangeEvent event) {
    Object[] listeners = this.listenerList.getListenerList() ;
    for (int i = listeners.length - 2 ; i >= 0 ; i -= 2) {
      if (listeners[i] == IAnnotationChangeListener.class) {
        ((IAnnotationChangeListener) listeners[i + 1]).annotationChanged(event) ;
      }
    }
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Indicates if some other object is "equal to" this one.</en>
   *
   * @param obj <zh>ҪȽϵĶ</zh>
   * <en>the reference object with which to compare.</en>
   * @return <zh>ȷtrue,򷵻false.</zh>
   * <en>true if this object is the same as the obj argument; false otherwise.</en>
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true ;
    }

    if (!(obj instanceof AbstractAnnotation)) {
      return false ;
    }

    AbstractAnnotation that = (AbstractAnnotation) obj ;
    if (!EqualsUtil.equals(this.tooltipText, that.tooltipText)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    int hashcode = super.hashCode() ;

    return hashcode * 37 + HashcodeUtil.hashCode(this.tooltipText) ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    AbstractAnnotation clone = (AbstractAnnotation)super.clone() ;
    clone.listenerList = new EventListenerList() ;

    return clone ;
  }

}
