package com.ijchart.xychart.plot ;

import java.awt.BasicStroke ;
import java.awt.Color ;
import java.awt.Font ;
import java.awt.GradientPaint ;
import java.awt.Graphics2D ;
import java.awt.Image ;
import java.awt.Paint ;
import java.awt.Stroke ;
import java.awt.geom.Point2D ;
import java.awt.geom.Rectangle2D ;
import java.io.IOException ;
import java.io.ObjectInputStream ;
import java.io.ObjectOutputStream ;
import java.io.Serializable ;
import java.util.Collection ;
import java.util.Collections ;
import java.util.List ;

import javax.swing.event.EventListenerList ;

import com.ijchart.xychart.chartinfo.ChartShapeCollection ;
import com.ijchart.xychart.chartinfo.PlotShapeCollection ;
import com.ijchart.xychart.constant.AxisLocation ;
import com.ijchart.xychart.constant.ChartChangeEventType ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.HorizontalAlignment ;
import com.ijchart.xychart.constant.PlotType ;
import com.ijchart.xychart.constant.RenderDrawOrder ;
import com.ijchart.xychart.constant.VerticalAlignment ;
import com.ijchart.xychart.event.AxisChangeEvent ;
import com.ijchart.xychart.event.DatasetChangeEvent ;
import com.ijchart.xychart.event.PlotChangeEvent ;
import com.ijchart.xychart.event.RenderChangeEvent ;
import com.ijchart.xychart.listener.IAxisChangeListener ;
import com.ijchart.xychart.listener.IDatasetChangeListener ;
import com.ijchart.xychart.listener.IPlotChangeListener ;
import com.ijchart.xychart.listener.IRenderChangeListener ;
import com.ijchart.xychart.shape.RectangleImageChartShape ;
import com.ijchart.xychart.shape.RectangleTextChartShape ;
import com.ijchart.xychart.ui.Align ;
import com.ijchart.xychart.ui.RectangleEdge ;
import com.ijchart.xychart.ui.RectangleInsets ;
import com.ijchart.xychart.ui.supplier.DefaultDrawingSupplier ;
import com.ijchart.xychart.ui.supplier.IDrawingSupplier ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;
import com.ijchart.xychart.util.SerialUtil ;
import com.ijchart.xychart.util.StringUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>plotĻ,{@link com.ijchart.xychart.client.IJChart}ίлݵplot.</zh>
 * <en>The base class for all plots in IJChart.The {@link com.ijchart.xychart.client.IJChart} class
 * delegates the drawing of axes and data to the plot.This base class provides facilities common
 * to most plot types.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public abstract class AbstractPlot implements IAxisChangeListener, IDatasetChangeListener,
  IRenderChangeListener, IZoomable, Cloneable, Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = 7489360884354202222L ;

  /**
   * <zh>ȱʡհ.</zh>
   * <en>The default insets.</en>
   */
  public static final RectangleInsets DEFAULT_INSETS
    = new RectangleInsets(4.0, 8.0, 4.0, 10.0) ;

  /**
   * <zh>ȱʡ߿Stroke.</zh>
   * <en>The default outline stroke.</en>
   */
  public static final Stroke DEFAULT_BORDER_STROKE = new BasicStroke(0.5f) ;

  /**
   * <zh>ȱʡ߿Paint.</zh>
   * <en>The default border color.</en>
   */
  public static final Paint DEFAULT_BORDER_PAINT = Color.gray ;

  /**
   * <zh>ȱʡǰalpha͸ֵ.</zh>
   * <en>The default foreground alpha transparency.</en>
   */
  public static final float DEFAULT_FOREGROUND_ALPHA = 1.0f ;

  /**
   * <zh>ȱʡalpha͸ֵ.</zh>
   * <en>The default background alpha transparency.</en>
   */
  public static final float DEFAULT_BACKGROUND_ALPHA = 1.0f ;

  /**
   * <zh>ȱʡɫ.</zh>
   * <en>The default background color. </en>
   */
  public static final Paint DEFAULT_BACKGROUND_PAINT = Color.white ;

  /**
   * <zh>plotС</zh>
   * <en>The default minimum width at which the plot should be drawn.</en>
   */
  public static final int DEFAULT_MINIMUM_WIDTH_TO_DRAW = 10 ;

  /**
   * <zh>plotС߶.</zh>
   * <en>The default minimum height at which the plot should be drawn.</en>
   */
  public static final int DEFAULT_MINIMUM_HEIGHT_TO_DRAW = 10 ;

  /**
   * <zh>ȱʡʱʾַ.</zh>
   * <en>The default message to display if no data is available.</en>
   */
  public static final String DEFAULT_NO_DATA_MESSAGE = "No data available" ;

  /**
   * <zh>ȱʡʱַ.</zh>
   * <en>The default font used to display the 'no data' message.</en>
   */
  public static final Font DEFAULT_NO_DATA_MESSAGE_FONT = new Font("SansSerif", Font.PLAIN, 12) ;

  /**
   * <zh>ȱʡʱַPaint.</zh>
   * <en>The paint used to draw the 'no data' message.</en>
   */
  public static final Paint DEFAULT_NO_DATA_MESSAGE_PAINT = Color.black ;

  /**
   * <zh>ȱʡǷ߿.</zh>
   * <en>A default flag that controls whether or not the plot border is drawn.</en>
   */
  public static final boolean DEFAULT_BORDER_VISIBLE = false ;

  /**
   * <zh>ȱʡıͼƬ뷽ʽ.</zh>
   * <en>The default background image alignment.</en>
   */
  public static final Align DEFAULT_BACKGROUND_IMAGE_ALIGNMENT = Align.FIT ;

  /**
   * <zh>ȱʡǷʾûʱϢ.</zh>
   * <en>A default flag that controls whether or not draw no data message
   * if no data is available.</en>
   */
  public static final boolean DEFAULT_DRAW_NO_DATA_MESSAGE = true ;

  /**
   * <zh>ṩͼҪһϵж.</zh>
   * <en>The drawing supplier for this render.</en>
   */
  public static final IDrawingSupplier DEFAULT_DRAWINT_SUPPLIER = new DefaultDrawingSupplier() ;

  /**
   * <zh>ȱʡrender˳.</zh>
   * <en>Default order for call renders.</en>
   */
  public static final RenderDrawOrder DEFAULT_RENDER_CALL_ORDER = RenderDrawOrder.FORWARD ;

  /**
   * <zh>ǷʾûʱϢ.</zh>
   * <en>A flag that controls whether or not draw no data message
   * if no data is available.</en>
   */
  public boolean drawNoDataMessage ;

  /**
   * <zh>ʱʾַ.</zh>
   * <en>The message to display if no data is available. </en>
   */
  private String noDataMessage ;

  /**
   * <zh>ȱʡʱַ.</zh>
   * <en>The font used to display the 'no data' message.</en>
   */
  private Font noDataMessageFont ;

  /**
   * <zh>ȱʡʱַPaint</zh>
   * <en> The paint used to draw the 'no data' message.</en>
   */
  private transient Paint noDataMessagePaint ;

  /**
   * <zh>plotܱߵĿհ.</zh>
   * <en>Amount of blank space around the plot area. </en>
   */
  private RectangleInsets insets ;

  /**
   * <zh>Ƿ߿.</zh>
   * <en>A flag that controls whether or not the plot border is drawn.</en>
   */
  private boolean borderVisible ;

  /**
   * <zh>߿Stroke.</zh>
   * <en>The Stroke used to draw a border around the plot.</en>
   */
  private transient Stroke borderStroke ;

  /**
   * <zh>߿Paint.</zh>
   * <en>The Paint used to draw a border around the plot.</en>
   */
  private transient Paint borderPaint ;

  /**
   * <zh>ƱPaint.</zh>
   * <en>An optional color used to fill the plot background. </en>
   */
  private transient Paint backgroundPaint ;

  /**
   * <zh>ͼƬ.</zh>
   * <en>An optional image for the plot background. </en>
   */
  private transient Image backgroundImage ; // not currently serialized

  /**
   * <zh>ͼƬ뷽ʽ.</zh>
   * <en>The alignment for the background image.</en>
   */
  private Align backgroundImageAlignment = null ;

  /**
   * <zh>ͼƬ͸alphaֵ.</zh>
   * <en>The alpha value used to draw the background image.</en>
   */
  private float backgroundImageAlpha ;

  /**
   * <zh>ǰ͸alphaֵ.</zh>
   * <en>The alpha-transparency for the plot. </en>
   */
  private float foregroundAlpha ;

  /**
   * <zh>͸alphaֵ.</zh>
   * <en>The alpha transparency for the background paint.</en>
   */
  private float backgroundAlpha ;

  /**
   * <zh>ṩͼҪһϵж.</zh>
   * <en>The drawing supplier for this render.</en>
   */
  private IDrawingSupplier drawingSupplier ;

  /**
   * <zh>¼.</zh>
   * <en>Storage for registered change listeners.</en>
   */
  private transient EventListenerList listenerList ;

  /**
   * <zh>renderĵ˳.</zh>
   * <en>the order for call renders.</en>
   */
  protected RenderDrawOrder renderingOrder ;

  /**
   * <zh>renderͼμϵļ.</zh>
   * <en>A list of render shape collection.</en>
   */
  protected List renderShapesList ;

  /**
   * <zh>.</zh>
   * <en>The data area.</en>
   */
  protected Rectangle2D dataArea ;

  /**
   * <zh>plotͼμ.</zh>
   * <en>Storage for plot information.</en>
   */
  protected PlotShapeCollection plotShapes ;

  /**
   * <zh>һµplot.</zh>
   * <en>Creates a new plot.</en>
   */
  protected AbstractPlot() {
    this.insets = DEFAULT_INSETS ;

    this.backgroundPaint = DEFAULT_BACKGROUND_PAINT ;
    this.backgroundAlpha = DEFAULT_BACKGROUND_ALPHA ;
    this.foregroundAlpha = DEFAULT_FOREGROUND_ALPHA ;

    this.backgroundImage = null ;
    this.backgroundImageAlignment = DEFAULT_BACKGROUND_IMAGE_ALIGNMENT ;
    this.backgroundImageAlpha = 0.5F ;

    this.borderVisible = DEFAULT_BORDER_VISIBLE ;
    this.borderStroke = DEFAULT_BORDER_STROKE ;
    this.borderPaint = DEFAULT_BORDER_PAINT ;

    this.drawNoDataMessage = DEFAULT_DRAW_NO_DATA_MESSAGE ;
    this.noDataMessage = DEFAULT_NO_DATA_MESSAGE ;
    this.noDataMessageFont = DEFAULT_NO_DATA_MESSAGE_FONT ;
    this.noDataMessagePaint = DEFAULT_NO_DATA_MESSAGE_PAINT ;

    this.drawingSupplier = DEFAULT_DRAWINT_SUPPLIER ;
    this.renderingOrder = DEFAULT_RENDER_CALL_ORDER ;
    this.listenerList = new EventListenerList() ;
    this.dataArea = null ;
  }

  /**
   * <zh>ʱʾַ.</zh>
   * <en>Returns the string that is displayed when the dataset is empty or null.</en>
   *
   * @return <zh>ַ.</zh>
   * <en>The 'no data' message (null possible).</en>
   */
  public String getNoDataMessage() {
    return this.noDataMessage ;
  }

  /**
   * <zh>ʱʾַ,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the message that is displayed when the dataset is empty or null, and sends
   * a {@link PlotChangeEvent} to all registered listeners.</en>
   *
   * @param message <zh>ַ.</zh>
   * <en>The message (null permitted).</en>
   */
  public void setNoDataMessage(String message) {
    this.noDataMessage = message ;
    firePlotChanged() ;
  }

  /**
   * <zh>ʱַ.</zh>
   * <en>Returns the font used to display the 'no data' message.</en>
   *
   * @return <zh>(Ϊnull).</zh>
   * <en>The font (never null).</en>
   */
  public Font getNoDataMessageFont() {
    return this.noDataMessageFont ;
  }

  /**
   * <zh>ʱַ,,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the font used to display the 'no data' message and sends a
   * {@link PlotChangeEvent} to all registered listeners.</en>
   *
   * @param font <zh> (Ϊnull).</zh>
   * <en>The font (null not permitted).</en>
   */
  public void setNoDataMessageFont(Font font) {
    if (font == null) {
      throw new IllegalArgumentException("Null 'font' argument.") ;
    }

    this.noDataMessageFont = font ;
    firePlotChanged() ;
  }

  /**
   * <zh>ʱַPaint.</zh>
   * <en>Returns the paint used to display the 'no data' message.</en>
   *
   * @return <zh>Paint (Ϊnull).</zh>
   * <en>The paint (never null).</en>
   */
  public Paint getNoDataMessagePaint() {
    return this.noDataMessagePaint ;
  }

  /**
   * <zh>ʱַPaint,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the paint used to display the 'no data' message and sends a
   * {@link PlotChangeEvent} to all registered listeners.</en>
   *
   * @param paint <zh>paint (Ϊnull).</zh>
   * <en>The paint (null not permitted).</en>
   */
  public void setNoDataMessagePaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument.") ;
    }

    this.noDataMessagePaint = paint ;
    firePlotChanged() ;
  }

  /**
   * <zh>plotַ.</zh>
   * <en>Returns a short string describing the plot type.</en>
   *
   * @return <zh>plot.</zh>
   * <en>The plot type.</en>
   */
  public abstract PlotType getPlotType() ;

  /**
   * <zh>plotܱߵĿհ.</zh>
   * <en>Returns the insets for the plot area.</en>
   *
   * @return <zh>հ(Ϊnull).</zh>
   * <en>The insets (never null).</en>
   */
  public RectangleInsets getInsets() {
    return this.insets ;
  }

  /**
   * <zh>plotܱߵĿհ,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the insets for the plot and sends a {@link PlotChangeEvent} to
   * all registered listeners.</en>
   *
   * @param insets <zh>հ (Ϊnull).</zh>
   * <en>The new insets (null not permitted).</en>
   */
  public void setInsets(RectangleInsets insets) {
    if (insets == null) {
      throw new IllegalArgumentException("Null 'insets' argument.") ;
    }

    if (!this.insets.equals(insets)) {
      this.insets = insets ;
      firePlotChanged() ;
    }
  }

  /**
   * <zh>ػƱPaint.</zh>
   * <en>Returns the background color of the plot area.</en>
   *
   * @return <zh>Paint (Ϊnull).</zh>
   * <en>The paint (possibly null).</en>
   */
  public Paint getBackgroundPaint() {
    return this.backgroundPaint ;
  }

  /**
   * <zh>ûƱPaint,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the background color of the plot area and sends a
   * {@link PlotChangeEvent} to all registered listeners.</en>
   *
   * @param paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (null permitted).</en>
   */
  public void setBackgroundPaint(Paint paint) {
    if (!(this.backgroundPaint == null && paint == null)) {
      this.backgroundPaint = paint ;
      firePlotChanged() ;
    }
  }

  /**
   * <zh>رalphaֵ.</zh>
   * <en>Returns the alpha transparency of the plot area background.</en>
   *
   * @return <zh>͸alphaֵ.</zh>
   * <en>The alpha transparency.</en>
   */
  public float getBackgroundAlpha() {
    return this.backgroundAlpha ;
  }

  /**
   * <zh>ñ͸alphaֵ,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the alpha transparency of the plot area background, and notifies
   * registered listeners that the plot has been modified.</en>
   *
   * @param alpha <zh>alphaֵ,ֵڱ[0.0,1.0].</zh>
   * <en>The new alpha value (in the range 0.0f to 1.0f).</en>
   */
  public void setBackgroundAlpha(float alpha) {
    if (alpha < 0.0f || alpha > 1.0f) {
      throw new IllegalArgumentException("The 'alpha' value must be in the range 0.0f to 1.0f.") ;
    }

    if (this.backgroundAlpha != alpha) {
      this.backgroundAlpha = alpha ;
      firePlotChanged() ;
    }
  }

  /**
   * <zh>رͼƬ.</zh>
   * <en>Returns the background image that is used to fill the plot's background area.</en>
   *
   * @return <zh>ͼƬ (Ϊnull).</zh>
   * <en>The image (possibly null).</en>
   */
  public Image getBackgroundImage() {
    return this.backgroundImage ;
  }

  /**
   * <zh>ñͼƬ,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the background image for the plot and sends a
   * {@link PlotChangeEvent} to all registered listeners.</en>
   *
   * @param image <zh>ͼƬ(Ϊnull).</zh>
   * <en>The image (null permitted).</en>
   */
  public void setBackgroundImage(Image image) {
    this.backgroundImage = image ;
    firePlotChanged() ;
  }

  /**
   * <zh>رͼƬĶ뷽ʽ.</zh>
   * <en>Returns the background image alignment. Alignment constants are defined
   * in the {@link Align} class.</en>
   *
   * @return <zh>뷽ʽ (Ϊnull).</zh>
   * <en>The alignment (never null).</en>
   */
  public Align getBackgroundImageAlignment() {
    return this.backgroundImageAlignment ;
  }

  /**
   * <zh>ñͼƬĶ뷽ʽ,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the alignment for the background image and sends a
   * {@link PlotChangeEvent} to all registered listeners.  Alignment options
   * are defined by the {@link Align} class.</en>
   *
   * @param alignment Align <zh>뷽ʽ.</zh>
   * <en>The alignment.</en>
   */
  public void setBackgroundImageAlignment(Align alignment) {
    if (!this.backgroundImageAlignment.equals(alignment)) {
      this.backgroundImageAlignment = alignment ;
      firePlotChanged() ;
    }
  }

  /**
   * <zh>ػƱͼƬalphaֵ.</zh>
   * <en>Returns the alpha transparency used to draw the background image.  This
   * is a value in the range 0.0f to 1.0f, where 0.0f is fully transparent
   * and 1.0f is fully opaque.</en>
   *
   * @return <zh>͸ȵalphaֵ.</zh>
   * <en>The alpha transparency.</en>
   */
  public float getBackgroundImageAlpha() {
    return this.backgroundImageAlpha ;
  }

  /**
   * <zh>ûƱͼƬalphaֵ,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the alpha transparency used when drawing the background image.</en>
   *
   * @param alpha <zh>alphaֵ,ڱ[0.0,1.0].</zh>
   * <en>The alpha transparency (in the range 0.0f to 1.0f, where 0.0f is fully transparent,
   * and 1.0f is fully opaque).</en>
   */
  public void setBackgroundImageAlpha(float alpha) {
    if (alpha < 0.0f || alpha > 1.0f) {
      throw new IllegalArgumentException(
        "The 'alpha' value must be in the range 0.0f to 1.0f.") ;
    }

    if (this.backgroundImageAlpha != alpha) {
      this.backgroundImageAlpha = alpha ;
      this.firePlotChanged() ;
    }
  }

  /**
   * <zh>߿Ƿɼ.</zh>
   * <en>Returns the flag that controls whether or not the plot border is drawn.</en>
   *
   * @return <zh>booleanֵ.</zh>
   * <en>The border visibility flag.</en>
   */
  public boolean isBorderVisible() {
    return this.borderVisible ;
  }

  /**
   * <zh>߿Ƿɼ,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the flag that controls whether or not the plot's border is
   * drawn, and sends a {@link PlotChangeEvent} to all registered listeners.</en>
   *
   * @param visible <zh>booleanֵ.</zh>
   * <en>The new flag value.</en>
   */
  public void setBorderVisible(boolean visible) {
    if (this.borderVisible != visible) {
      this.borderVisible = visible ;
      firePlotChanged() ;
    }
  }

  /**
   * <zh>߿Stroke.</zh>
   * <en>Returns the stroke used to outline the plot area.</en>
   *
   * @return <zh>Stroke (Ϊnull).</zh>
   * <en>The stroke (never null).</en>
   */
  public Stroke getBorderStroke() {
    return this.borderStroke ;
  }

  /**
   * <zh>߿Stroke,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the stroke used to outline the plot area and sends a
   * {@link PlotChangeEvent} to all registered listeners. </en>
   *
   * @param stroke <zh>Stroke(Ϊnull).</zh>
   * <en>The stroke (null not permitted).</en>
   */
  public void setBorderStroke(Stroke stroke) {
    if (stroke == null) {
      throw new IllegalArgumentException("Null 'stroke' argument.") ;
    }

    this.borderStroke = stroke ;
    firePlotChanged() ;
  }

  /**
   * <zh>߿Paint.</zh>
   * <en>Returns the color used to draw the border of the plot area.</en>
   *
   * @return <zh>Paint(Ϊnull).</zh>
   * <en>The color (never null).</en>
   */
  public Paint getBorderPaint() {
    return this.borderPaint ;
  }

  /**
   * <zh>߿Paint,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the paint used to draw the border of the plot area and sends a
   * {@link PlotChangeEvent} to all registered listeners.</en>
   *
   * @param paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (null not permitted).</en>
   */
  public void setBorderPaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument.") ;
    }

    this.borderPaint = paint ;
    firePlotChanged() ;
  }

  /**
   * <zh>ǰalphaֵ.</zh>
   * <en>Returns the alpha-transparency for the plot foreground.</en>
   *
   * @return <zh>alphaֵ.</zh>
   * <en>The alpha-transparency.</en>
   */
  public float getForegroundAlpha() {
    return this.foregroundAlpha ;
  }

  /**
   * <zh>ǰalphaֵ,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the alpha-transparency for the plot and sends a
   * {@link PlotChangeEvent} to all registered listeners.</en>
   *
   * @param alpha <zh>alphaֵ,ڱ[0.0,1.0].</zh>
   * <en>The new alpha transparency (in the range 0.0f to 1.0f, where 0.0f is fully transparent,
   * and 1.0f is fully opaque).</en>
   */
  public void setForegroundAlpha(float alpha) {
    if (alpha < 0.0f || alpha > 1.0f) {
      throw new IllegalArgumentException(
        "The 'alpha' value must be in the range 0.0f to 1.0f.") ;
    }

    if (this.foregroundAlpha != alpha) {
      this.foregroundAlpha = alpha ;
      firePlotChanged() ;
    }
  }

  /**
   * <zh>ǷʾûʱϢ.</zh>
   * <en>Returns the flag that controls whether or not draw no data message
   * if no data is available.</en>

   * @return boolean <zh>trueʾ,falseʾ.</zh>
   * <en>Returns true if draw no data message,otherwise return false.</en>
   */
  public boolean isDrawNoDataMessage() {
    return drawNoDataMessage ;
  }

  /**
   * <zh>ǷʾûʱϢ.</zh>
   * <en>A flag that controls whether or not draw no data message
   * if no data is available.</en>

   * @param flag boolean <zh>trueʾ,falseʾ.</zh>
   * <en>Returns true if draw no data message,otherwise return false.</en>
   */
  public void setDrawNoDataMessage(boolean flag) {
    if (this.drawNoDataMessage != flag) {
      this.drawNoDataMessage = flag ;
      firePlotChanged() ;
    }
  }

  /**
   * <zh>{@link IDrawingSupplier},ṩͼҪһϵж.</zh>
   * <en>Returns the drawing supplier for this render.</en>
   *
   * @return IDrawingSupplier <zh>IDrawingSupplier (Ϊnull).</zh>
   * <en>The drawing supplier for this render (never null).</en>
   */
  public IDrawingSupplier getDrawingSupplier() {
    return this.drawingSupplier ;
  }

  /**
   * <zh>{@link IDrawingSupplier},ṩͼҪһϵж,
   * м{@link RenderChangeEvent}¼.</zh>
   * <en>Sets the drawing supplier for this render,and
   * sends a {@link RenderChangeEvent} to all registered listeners.</en>
   *
   * @param supplier IDrawingSupplier <zh>IDrawingSupplier (Ϊnull).</zh>
   * <en>The drawing supplier for this render (null not permitted).</en>
   */
  public void setDrawingSupplier(IDrawingSupplier supplier) {
    if (supplier == null) {
      throw new IllegalArgumentException("Null 'supplier' argument.") ;
    }

    this.drawingSupplier = supplier ;
    firePlotChanged() ;
  }

  /**
   * <zh>ͼplot,Ҫʵִ˷.</zh>
   * <en>Draws the plot within the specified area.Subclasses need to provide an
   * implementation of this method, obviously.</en>
   *
   * @param g2 Graphics2D <zh>ͼ豸(ӡĻ) (Ϊnull).</zh>
   * <en>The graphics device (null not permitted).</en>
   * @param plotArea Rectangle2D <zh>plot (Ϊnull).</zh>
   * <en>The plot draw area (null not permitted).</en>
   * @param chartShapes ChartShapeCollection <zh>chartͼμ (Ϊnull).</zh>
   * <en>Collects information about the chart (null not permitted).</en>
   */
  public abstract void draw(Graphics2D g2, Rectangle2D plotArea,
                            ChartShapeCollection chartShapes) ;

  /**
   * <zh>ʮ,Ҫʵִ˷.</zh>
   * <en>Draws the crosshair.</en>
   *
   * @param mouseClickPoint Point2D <zh>ʱ (Ϊnull).</zh>
   * <en>The mouse click point (null permitted).</en>
   *
   * @return boolean <zh>Ƿʮ߷˱仯.</zh>
   * <en>Returns whether or not crosshair is changed.</en>
   */
  public abstract boolean drawCrosshair(Point2D mouseClickPoint) ;

  /**
   * <zh>ʮ߶Ӧĵ.</zh>
   * <en>Returns the crosshair point.</en>
   *
   * @param mouseClickPoint Point2D<zh>ʱ (Ϊnull).</zh>
   * <en>The mouse click point (null permitted).</en>
   *
   * @return Point2D <zh> (Ϊnull).</zh>
   * <en>The point (maybe null).</en>
   */
  public Point2D getCrosshairPoint(Point2D mouseClickPoint) {
    return mouseClickPoint ;
  }

  /**
   * <zh>Ʊɫ,ƱPaintGradientPaint,gradientĻƷͼĻƷһ.</zh>
   * <en>Fills the specified area with the background paint.  If the background
   * paint is an instance of GradientPaint, the gradient will
   * run in the direction suggested by the plot's orientation.</en>
   *
   * @param plotArea <zh> (Ϊnull).</zh>
   * <en>The background draw area (null not permitted).</en>
   * @param orientation ChartOrientation <zh>ͼƷ.</zh>
   * <en>The chart draw orientation.</en>
   * @param shapeCollection RenderShapeCollection <zh>plotͼμ (Ϊnull).</zh>
   * <en>Collects information about the plot (null not permitted).</en>
   */
  protected void drawBackground(Rectangle2D plotArea, ChartOrientation orientation,
                                PlotShapeCollection shapeCollection) {
    if (plotArea == null) {
      throw new IllegalArgumentException("Null 'plotArea' argument.") ;
    }
    if (orientation == null) {
      throw new IllegalArgumentException("Null 'orientation' argument.") ;
    }
    if (shapeCollection == null) {
      throw new IllegalArgumentException("Null 'shapeCollection' argument.") ;
    }

    if (this.backgroundPaint == null) {
      return ;
    }

    Paint p = this.backgroundPaint ;
    if (p instanceof GradientPaint) {
      GradientPaint gp = (GradientPaint) p ;
      if (orientation.equals(ChartOrientation.VERTICAL)) {
        p = new GradientPaint((float) plotArea.getCenterX(),
                              (float) plotArea.getMaxY(), gp.getColor1(),
                              (float) plotArea.getCenterX(), (float) plotArea.getMinY(),
                              gp.getColor2()) ;
      } else if (orientation.equals(ChartOrientation.HORIZONTAL)) {
        p = new GradientPaint((float) plotArea.getMinX(),
                              (float) plotArea.getCenterY(), gp.getColor1(),
                              (float) plotArea.getMaxX(), (float) plotArea.getCenterY(),
                              gp.getColor2()) ;
      }
    }

    //ɫͼ,֮Լȥ0.5Ϊֱɫһ
    Rectangle2D drawArea = new Rectangle2D.Double(plotArea.getX(), plotArea.getY() - 0.5,
                                                  plotArea.getWidth(), plotArea.getHeight() + 1) ;
    RectangleTextChartShape shape = new RectangleTextChartShape(drawArea, null) ;
    shape.setDrawBorder(false) ;
    shape.setFillPaint(p) ;
    shape.setFillShape(true) ;
    shape.setFillAlpha(this.backgroundAlpha) ;

    shapeCollection.setBackgroundShape(shape) ;
  }

  /**
   * <zh>ָڻƱͼƬ.</zh>
   * <en>Draws the background image (if there is one) aligned within the specified area.</en>
   *
   * @param plotArea <zh> (Ϊnull).</zh>
   * <en>The background draw area (null not permitted).</en>
   * @param shapeCollection RenderShapeCollection <zh>plotͼμ (Ϊnull).</zh>
   * <en>Collects information about the plot (null not permitted).</en>
   */
  protected void drawBackgroundImage(Rectangle2D plotArea,
                                     PlotShapeCollection shapeCollection) {
    if (plotArea == null) {
      throw new IllegalArgumentException("Null 'plotArea' argument.") ;
    }
    if (shapeCollection == null) {
      throw new IllegalArgumentException("Null 'shapeCollection' argument.") ;
    }

    if (this.backgroundImage != null) {
      Rectangle2D dest = new Rectangle2D.Double(0.0, 0.0, this.backgroundImage.getWidth(null),
                                                this.backgroundImage.getHeight(null)) ;
      Rectangle2D imageRect = Align.align(dest, plotArea, this.backgroundImageAlignment) ;
      imageRect.setRect(imageRect.getX(), imageRect.getY(), imageRect.getWidth() + 1,
                        imageRect.getHeight() + 1) ;

      //ͼƬͼ
      RectangleImageChartShape shape = new RectangleImageChartShape(imageRect, this.backgroundImage) ;
      shape.setDrawBorder(false) ;
      shape.setFillShape(false) ; //ͼ
      shape.setImageAlpha(this.backgroundImageAlpha) ;

      shapeCollection.setBackgroundImageShape(shape) ;
    }
  }

  /**
   * <zh>߿.</zh>
   * <en>Draws the plot border.</en>
   *
   * @param plotArea <zh> (Ϊnull).</zh>
   * <en>The background draw area (null not permitted).</en>
   * @param shapeCollection RenderShapeCollection <zh>plotͼμ (Ϊnull).</zh>
   * <en>Collects information about the plot (null not permitted).</en>
   */
  protected void drawBorder(Rectangle2D plotArea, PlotShapeCollection shapeCollection) {
    if (plotArea == null) {
      throw new IllegalArgumentException("Null 'plotArea' argument.") ;
    }
    if (shapeCollection == null) {
      throw new IllegalArgumentException("Null 'shapeCollection' argument.") ;
    }

    if (this.borderVisible && this.borderStroke != null && this.borderPaint != null) {
      //ͼƬͼ
      RectangleTextChartShape shape = new RectangleTextChartShape(plotArea, null) ;
      shape.setDrawBorder(true) ;
      shape.setBorderPaint(this.borderPaint) ;
      shape.setBorderStroke(this.borderStroke) ;
      shape.setFillShape(false) ;

      shapeCollection.setBorderShape(shape) ;
    }
  }

  /**
   * <zh>ݼûʱʾϢ.</zh>
   * <en>Draws a message to state that there is no data to plot.</en>
   *
   * @param plotArea <zh> (Ϊnull).</zh>
   * <en>The background draw area (null not permitted).</en>
   * @param shapeCollection RenderShapeCollection <zh>plotͼμ (Ϊnull).</zh>
   * <en>Collects information about the plot (null not permitted).</en>
   */
  protected void drawNoDataMessage(Rectangle2D plotArea, PlotShapeCollection shapeCollection) {
    if (plotArea == null) {
      throw new IllegalArgumentException("Null 'plotArea' argument.") ;
    }
    if (shapeCollection == null) {
      throw new IllegalArgumentException("Null 'shapeCollection' argument.") ;
    }

    if (!StringUtil.isEmpty(this.noDataMessage)) {
      //ݼΪʱʾַͼ
      RectangleTextChartShape shape = new RectangleTextChartShape(plotArea, this.noDataMessage) ;
      shape.setDrawBorder(false) ;
      shape.setFillShape(false) ;
      shape.setTextFont(this.noDataMessageFont) ;
      shape.setTextPaint(this.noDataMessagePaint) ;
      shape.setTextHorizontalAlign(HorizontalAlignment.CENTER) ;
      shape.setTextVerticalAlign(VerticalAlignment.CENTER) ;

      shapeCollection.setNoDataMessageShape(shape) ;
    }
  }

  /**
   * <zh>renderĵ˳.</zh>
   * <en>Returns the order for call renders.</en>
   *
   * @return RenderDrawOrder <zh>renderĵ˳ (Ϊnull).</zh>
   * <en>The order for call renders (never null).</en>
   */
  public RenderDrawOrder getRenderDrawOrder() {
    return this.renderingOrder ;
  }

  /**
   * <zh>renderĵ˳,м{@link PlotChangeEvent}¼.</zh>
   * <en>Sets the renders call order and sends a {@link PlotChangeEvent} to all
   * registered listeners.</en>
   *
   * @param order RenderDrawOrder <zh>˳ (Ϊnull).</zh>
   * <en>The rendering order (null not permitted).</en>
   */
  public void setRenderDrawOrder(RenderDrawOrder order) {
    if (order == null) {
      throw new IllegalArgumentException("Null 'order' argument.") ;
    }

    if (!this.renderingOrder.equals(order)) {
      this.renderingOrder = order ;
      notifyListeners(new PlotChangeEvent(this)) ;
    }
  }

  /**
   * <zh>renderͼμϵļ.</zh>
   * <en>Returns the list of render shape collection.</en>
   *
   * @return Collection <zh>,˼ϲ޸.</zh>
   * <en>The collection and unmodifiable.</en>
   */
  public Collection getRenderShapesList() {
    return Collections.unmodifiableCollection(this.renderShapesList) ;
  }

  /**
   * <zh>render,˼ϲ޸.</zh>
   * <en>Returns the collection of renders,and this collection is unmodifiable.</en>
   *
   * @return Collection <zh>render.</zh>
   * <en>The collection of renders.</en>
   */
  public abstract Collection getRenders() ;

  /**
   * <zh>.</zh>
   * <en>The data area.</en>
   *
   * @return dataArea Rectangle2D <zh>.</zh>
   * <en>The area.</en>
   */
  public Rectangle2D getDataArea() {
    return this.dataArea ;
  }

  /**
   * <zh>domainǷ.</zh>
   * <en>Returns true if the plot's domain axis is zoomable, and false otherwise.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isDomainZoomable() {
    return true ;
  }

  /**
   * <zh>valueǷ.</zh>
   * <en>Returns true if the plot's value is zoomable, and false otherwise.</en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isValueZoomable() {
    return true ;
  }

  /**
   * <zh>еplot.</zh>
   * <en>Returns all sub plots.</en>
   *
   * @return Collection <zh>plot󼯺,˶󲻿޸.</zh>
   * <en>The collection of sub plots (unmodifiable).</en>
   */
  public Collection getSubPlots() {
    return null ;
  }

  /**
   * <zh>domain.</zh>
   * <en>Zoom domain axes.</en>
   *
   * @param selectArea Rectangle2D <zh>ѡ,Ϊnullʾָǰ״̬.</zh>
   * <en>The selected area,null to restore chart.</en>
   * @param dataArea Rectangle2D <zh>,Ϊnullʾָǰ״̬.</zh>
   * <en>The data area,null to restore chart.</en>
   * @param source Point2D <zh>Դ (Ϊnull).</zh>
   * <en>The source mouse point (null not permitted).</en>
   */
  public void zoomDomainAxes(Rectangle2D selectArea, Rectangle2D dataArea, Point2D source) {
    //No action
  }

  /**
   * <zh>value.</zh>
   * <en>Zoom value axes.</en>
   *
   * @param selectArea Rectangle2D <zh>ѡ,Ϊnullʾָǰ״̬.</zh>
   * <en>The selected area,null to restore chart.</en>
   * @param dataArea Rectangle2D <zh>,Ϊnullʾָǰ״̬.</zh>
   * <en>The data area,null to restore chart.</en>
   * @param source Point2D <zh>Դ (Ϊnull).</zh>
   * <en>The source mouse point (null not permitted).</en>
   */
  public void zoomValueAxes(Rectangle2D selectArea, Rectangle2D dataArea, Point2D source) {
    //No action
  }

  /**
   * Handles a 'click' on the plot.  Since the plot does not maintain any
   * information about where it has been drawn, the plot rendering info is
   * supplied as an argument.
   *
   * @param x  the x coordinate (in Java2D space).
   * @param y  the y coordinate (in Java2D space).
   * @param shapeCollection  an object containing information about the dimensions of
   *              the plot.
   */
  public void handleClick(int x, int y, ChartShapeCollection shapeCollection) {
    //no action
  }

  /**
   * <zh>յ仯¼,¼ݸеļ.</zh>
   * <en>Receives notification of a change to one of the plot's axes.The plot reacts by passing on
   *  a axis change event to all registered listeners.</en>
   *
   * @param event <zh>仯¼</zh>
   * <en>Information about the event (not used here).</en>
   */
  public void axisChanged(AxisChangeEvent event) {
    notifyListeners(new PlotChangeEvent(this)) ;
  }

  /**
   * <zh>ݼ仯¼,¼ݸеļ.</zh>
   * <en>Receives notification of a change to the plot's dataset.
   * The plot reacts by passing on a plot change event to all registered listeners.</en>
   *
   * @param event <zh>ݼ仯¼.</zh>
   * <en>Information about the event (not used here).</en>
   */
  public void datasetChanged(DatasetChangeEvent event) {
    PlotChangeEvent newEvent = new PlotChangeEvent(this) ;
    newEvent.setType(ChartChangeEventType.DATASET_UPDATED) ;
    notifyListeners(newEvent) ;
  }

  /**
   * <zh>һ.</zh>
   * <en>Registers an object for notification of changes to the plot.</en>
   *
   * @param listener <zh> (Ϊnull).</zh>
   * <en>The object to be registered (null not permitted).</en>
   */
  public void addChangeListener(IPlotChangeListener listener) {
    if (listener == null) {
      throw new IllegalArgumentException("Null 'listener' argument.") ;
    }

    this.listenerList.add(IPlotChangeListener.class, listener) ;
  }

  /**
   * <zh>ɾһ.</zh>
   * <en>Unregisters an object for notification of changes to the plot.</en>
   *
   * @param listener <zh>Ҫɾļ (Ϊnull).</zh>
   * <en>The object to be unregistered (null not permitted).</en>
   */
  public void removeChangeListener(IPlotChangeListener listener) {
    if (listener == null) {
      throw new IllegalArgumentException("Null 'listener' argument.") ;
    }

    this.listenerList.remove(IPlotChangeListener.class, listener) ;
  }

  /**
   * <zh>ͨмplot˱仯.</zh>
   * <en>Notifies all registered listeners that the plot has changed.</en>
   */
  protected void firePlotChanged() {
    notifyListeners(new PlotChangeEvent(this)) ;
  }

  /**
   * <zh>յrender仯¼.</zh>
   * <en>Receives notification of a render change event.</en>
   *
   * @param event RenderChangeEvent <zh>render仯¼.</zh>
   * <en>The render change event.</en>
   */
  public void renderChanged(RenderChangeEvent event) {
    notifyListeners(new PlotChangeEvent(this)) ;
  }

  /**
   * <zh>֪ͨ{@link PlotChangeEvent}¼.</zh>
   * <en>Notifies all registered listeners that the plot has been modified.</en>
   *
   * @param event <zh>{@link PlotChangeEvent}¼ (Ϊnull).</zh>
   * <en>Information about the change event (null not permitted).</en>
   */
  protected void notifyListeners(PlotChangeEvent event) {
    if (event == null) {
      throw new IllegalArgumentException("Null 'event' argument.") ;
    }

    Object[] listeners = this.listenerList.getListenerList() ;
    for (int i = listeners.length - 2 ; i >= 0 ; i -= 2) {
      if (listeners[i] == IPlotChangeListener.class) {
        ((IPlotChangeListener) listeners[i + 1]).plotChanged(event) ;
      }
    }
  }

  /**
   * <zh>chartķdomainλȷλ.</zh>
   * <en>Resolves a domain axis location for a given chart orientation.</en>
   *
   * @param location <zh>λ (Ϊnull).</zh>
   * <en>The location (null not permitted).</en>
   * @param orientation <zh>ͼķ (Ϊnull).</zh>
   * The chart orientation (null not permitted).
   *
   * @return <zh>λ (Ϊnull).</zh>
   * <en>The edge (never null).</en>
   */
  public static RectangleEdge resolveDomainAxisLocation(AxisLocation location,
                                                        ChartOrientation orientation) {
    if (location == null) {
      throw new IllegalArgumentException("Null 'location' argument.") ;
    }
    if (orientation == null) {
      throw new IllegalArgumentException("Null 'orientation' argument.") ;
    }

    RectangleEdge result = null ;

    if (location.equals(AxisLocation.TOP_OR_RIGHT)) {
      if (orientation.equals(ChartOrientation.HORIZONTAL)) {
        result = RectangleEdge.RIGHT ;
      } else if (orientation.equals(ChartOrientation.VERTICAL)) {
        result = RectangleEdge.TOP ;
      }
    } else if (location.equals(AxisLocation.TOP_OR_LEFT)) {
      if (orientation.equals(ChartOrientation.HORIZONTAL)) {
        result = RectangleEdge.LEFT ;
      } else if (orientation.equals(ChartOrientation.VERTICAL)) {
        result = RectangleEdge.TOP ;
      }
    } else if (location.equals(AxisLocation.BOTTOM_OR_RIGHT)) {
      if (orientation.equals(ChartOrientation.HORIZONTAL)) {
        result = RectangleEdge.RIGHT ;
      } else if (orientation.equals(ChartOrientation.VERTICAL)) {
        result = RectangleEdge.BOTTOM ;
      }
    } else if (location.equals(AxisLocation.BOTTOM_OR_LEFT)) {
      if (orientation.equals(ChartOrientation.HORIZONTAL)) {
        result = RectangleEdge.LEFT ;
      } else if (orientation.equals(ChartOrientation.VERTICAL)) {
        result = RectangleEdge.BOTTOM ;
      }
    }

    return result ;

  }

  /**
   * <zh>chartķvalueλȷλ.</zh>
   * <en>Resolves a value axis location for a given chart orientation.</en>
   *
   * @param location <zh>λ (Ϊnull).</zh>
   * <en>The location (null not permitted).</en>
   * @param orientation <zh>ͼķ (Ϊnull).</zh>
   * The chart orientation (null not permitted).
   *
   * @return <zh>λ (Ϊnull).</zh>
   * <en>The edge (never null).</en>
   */

  public static RectangleEdge resolveValueAxisLocation(AxisLocation location,
                                                       ChartOrientation orientation) {

    if (location == null) {
      throw new IllegalArgumentException("Null 'location' argument.") ;
    }
    if (orientation == null) {
      throw new IllegalArgumentException("Null 'orientation' argument.") ;
    }

    RectangleEdge result = null ;

    if (location.equals(AxisLocation.TOP_OR_RIGHT)) {
      if (orientation.equals(ChartOrientation.HORIZONTAL)) {
        result = RectangleEdge.TOP ;
      } else if (orientation.equals(ChartOrientation.VERTICAL)) {
        result = RectangleEdge.RIGHT ;
      }
    } else if (location.equals(AxisLocation.TOP_OR_LEFT)) {
      if (orientation.equals(ChartOrientation.HORIZONTAL)) {
        result = RectangleEdge.TOP ;
      } else if (orientation.equals(ChartOrientation.VERTICAL)) {
        result = RectangleEdge.LEFT ;
      }
    } else if (location.equals(AxisLocation.BOTTOM_OR_RIGHT)) {
      if (orientation.equals(ChartOrientation.HORIZONTAL)) {
        result = RectangleEdge.BOTTOM ;
      } else if (orientation.equals(ChartOrientation.VERTICAL)) {
        result = RectangleEdge.RIGHT ;
      }
    } else if (location.equals(AxisLocation.BOTTOM_OR_LEFT)) {
      if (orientation.equals(ChartOrientation.HORIZONTAL)) {
        result = RectangleEdge.BOTTOM ;
      } else if (orientation.equals(ChartOrientation.VERTICAL)) {
        result = RectangleEdge.LEFT ;
      }
    }

    return result ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Tests this plot for equality with another object.</en>
   *
   * @param obj  the object (null permitted).
   *
   * @return true or false.
   */
  public boolean equals(Object obj) {
    if (obj == this) {
      return true ;
    }
    if (!(obj instanceof AbstractPlot)) {
      return false ;
    }

    AbstractPlot that = (AbstractPlot) obj ;

    if (!EqualsUtil.equals(this.noDataMessage, that.noDataMessage)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.noDataMessageFont, that.noDataMessageFont)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.noDataMessagePaint, that.noDataMessagePaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.insets, that.insets)) {
      return false ;
    }
    if (this.borderVisible != that.borderVisible) {
      return false ;
    }
    if (!EqualsUtil.equals(this.borderStroke, that.borderStroke)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.borderPaint, that.borderPaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.backgroundPaint, that.backgroundPaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.backgroundImage, that.backgroundImage)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.backgroundImageAlignment, that.backgroundImageAlignment)) {
      return false ;
    }
    if (this.backgroundImageAlpha != that.backgroundImageAlpha) {
      return false ;
    }
    if (this.foregroundAlpha != that.foregroundAlpha) {
      return false ;
    }
    if (this.backgroundAlpha != that.backgroundAlpha) {
      return false ;
    }
    if (!EqualsUtil.equals(this.drawingSupplier, that.drawingSupplier)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.renderingOrder, that.renderingOrder)) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    int hashCode = super.hashCode() ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.noDataMessage) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.noDataMessageFont) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.noDataMessagePaint) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.insets) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.borderVisible) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.borderStroke) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.borderPaint) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.backgroundPaint) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.backgroundImage) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.backgroundImageAlignment) ;
    hashCode = 37 * hashCode + (int) (100 * this.backgroundImageAlpha) ;
    hashCode = 37 * hashCode + (int) (100 * this.foregroundAlpha) ;
    hashCode = 37 * hashCode + (int) (100 * this.backgroundAlpha) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.drawingSupplier) ;
    hashCode = 37 * hashCode + HashcodeUtil.hashCode(this.renderingOrder) ;

    return hashCode ;
  }

  /**
   * <zh>¡.</zh>
   * <en>Creates a clone of the plot.</en>
   *
   * @return A clone.
   *
   * @throws CloneNotSupportedException if some component of the plot does not
   *         support cloning.
   */
  public Object clone() throws CloneNotSupportedException {
    AbstractPlot clone = (AbstractPlot)super.clone() ;

    clone.listenerList = new EventListenerList() ;
    return clone ;

  }

  /**
   * Provides serialization support.
   *
   * @param stream  the output stream.
   *
   * @throws IOException  if there is an I/O error.
   */
  private void writeObject(ObjectOutputStream stream) throws IOException {
    stream.defaultWriteObject() ;
    SerialUtil.writePaint(stream, this.noDataMessagePaint) ;
    SerialUtil.writeStroke(stream, this.borderStroke) ;
    SerialUtil.writePaint(stream, this.borderPaint) ;
    //dont write backgroundImage /** @todo  */
    SerialUtil.writePaint(stream, this.backgroundPaint) ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the input stream.
   *
   * @throws IOException  if there is an I/O error.
   * @throws ClassNotFoundException  if there is a classpath problem.
   */
  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject() ;
    this.noDataMessagePaint = SerialUtil.readPaint(stream) ;
    this.borderStroke = SerialUtil.readStroke(stream) ;
    this.borderPaint = SerialUtil.readPaint(stream) ;
    //dont read backgroundImage
    this.backgroundPaint = SerialUtil.readPaint(stream) ;
    this.listenerList = new EventListenerList() ;
  }

  /**
   * <zh>listΪindexԪصֵ.</zh>
   * <en>Replaces the element at the specified position in this list with
   * the specified element.</en>
   *
   * @param list List <zh>list (Ϊnull).</zh>
   * <en>The list (null not permitted).</en>
   * @param index int <zh>ֵ,ڵ0.</zh>
   * <en>The dataset index (zero-based).</en>
   * @param obj Object <zh> (Ϊnull).</zh>
   * <en>The object (null permitted).</en>
   */
  protected void setListObject(List list, int index, Object obj) {
    if (list == null) {
      throw new IllegalArgumentException("Null 'list' argument.") ;
    }
    if (index < 0) {
      throw new IllegalArgumentException("Must be 'index'>=0.") ;
    }

    int size = list.size() ;
    if (index > size - 1) {
      for (int i = 0 ; i < index - (size - 1) ; i++) {
        list.add(null) ;
      }
    }

    list.set(index, obj) ;
  }

}
